#!/usr/bin/perl -I/usr/sausalito/perl
# $Id: backup_restore.pl

# This handler is triggered whenever a user initiates a restore a VPS backup
# through the GUI.

$whatami = "handler";

use CCE;
my $cce = new CCE('Domain' => 'base-vserver');
use POSIX qw(isalpha);
use Socket;
use Sys::Hostname;
use File::Spec;
use List::MoreUtils qw(uniq);
use Net::SFTP::Foreign;
use File::Basename;
use POSIX qw(strftime);
use DateTime;
use Data::Dumper;

# Helper to redirect SSH errors to /dev/null:
open $ssherr, '>', '/dev/null'; # or die "Unable to open /dev/null";

use File::Remote qw(:replace);       # special :replace tag
setrsh("/usr/bin/ssh");
setrcp("/usr/bin/scp");

# Debugging switch:
$DEBUG = "1";
if ($DEBUG) {
    use Sys::Syslog qw( :DEFAULT setlogsock);
}

if ($whatami eq "handler") {
    $cce->connectfd();
}
else {
    $cce->connectuds();
}

# If this is a secondary in a cluster, we can save us the bother:
if (!-d "/vz/private") {
    $cce->bye('SUCCESS');
    exit 0;
}

$event = $cce->event_object();
$event_new = $cce->event_new();
$event_old = $cce->event_old();

&debug_msg("INFO: Passed cluster check \n");
&debug_msg("INFO: event: $event->{'OID'} \n");
&debug_msg("INFO: event_new: $event_new->{'OID'} \n");
&debug_msg("INFO: event_old: $event_old->{'OID'} \n");

# Get settings out of CODB to determine the target of the remote backup (if any):
@oids = $cce->find('VserverMasterNode');
($ok, $solbackupValues) = $cce->get($oids[0]);
if ($ok) {
    # Define values:
    $vzdump_enabled             = $solbackupValues->{'vzdump_enabled'};
    $rsync_enabled              = $solbackupValues->{'rsync_enabled'};
    $MN_RemoteBackupEnabled     = $solbackupValues->{'MN_RemoteBackupEnabled'};
    $MN_LocalBackupEnabled      = $solbackupValues->{'MN_LocalBackupEnabled'};
    $MN_BackupLocalLocation     = $solbackupValues->{'MN_BackupLocalLocation'};
    $targethost                 = $solbackupValues->{'MN_BackupRemoteTargetHost'};
    $targetip                   = $solbackupValues->{'MN_BackupRemoteTargetIP'};
    $path                       = $solbackupValues->{'MN_BackupRemoteLocation'};
    $user                       = $solbackupValues->{'MN_BackupRemoteTargetUser'};
    $pass                       = $solbackupValues->{'MN_BackupRemoteTargetUserPass'};
    $port                       = $solbackupValues->{'MN_BackupRemoteTargetPort'};
    $limit                      = $solbackupValues->{'RsyncSpeedLimit'};
    $my_maxfiles                = $solbackupValues->{'vzdump_maxfiles'};
    $kvm_support                = $solbackupValues->{'kvm_support'};

    # Assemble paths:
    $local_location_vzdump = $solbackupValues->{'MN_BackupLocalLocation'} . "/vzbackup";
    $local_location_vzdump =~ s|//|/|g;
    $local_location_vzdump =~ s|/$||g;

    $remote_location_vzdump = $solbackupValues->{'MN_BackupRemoteLocation'};
    $remote_location_vzdump =~ s|//|/|g;
    $remote_location_vzdump =~ s|/$||g;

    $local_location_rsync = $solbackupValues->{'MN_BackupLocalLocation'} . "/rsync";
    $local_location_rsync =~ s|//|/|g;
    $local_location_rsync =~ s|/$||g;

    $remote_location_rsync = $solbackupValues->{'MN_BackupRemoteLocation'};
    $remote_location_rsync =~ s|//|/|g; 
    $remote_location_rsync =~ s|/$||g; 
}

# Get the data of the Backup Object we want to restore:
($ok, $Backup) = $cce->get($event->{'OID'});
if ($ok) {
    if (($Backup->{'location'} eq "local") && ($Backup->{'origin'} eq "rsync") && (-d $Backup->{'storage'}) && ($Backup->{'targetvpsid'})) {
        $vpsid = $Backup->{'vpsid'};
        if ($Backup->{'virttype'} eq "openvz") {
            &debug_msg("INFO: OID $event->{'OID'} containing a local OpenVZ rsync backup ($vpsid) at $Backup->{'storage'} is going to be restored. \n");
            #
            ### Local Backup + Simfs:
            #
            if ((-d "$Backup->{'storage'}" ) && (-e "$Backup->{'storage'}/etc/vzdump/vps.conf" )) {
                $vps_config_file = "$Backup->{'storage'}/etc/vzdump/vps.conf";
                &debug_msg("INFO: Found $Backup->{'storage'} containing a local rsync backup and its config file $vps_config_file. \n");
                if (($Backup->{'virttype'} eq "openvz") && (-d "/vz/private/$Backup->{'targetvpsid'}")) {
                    &debug_msg("INFO: Stopping old VPS $Backup->{'targetvpsid'} in preparation for the restore. \n");
                    system("/usr/sbin/vzctl stop $Backup->{'targetvpsid'}");
                    &debug_msg("INFO: Destroying old VPS $Backup->{'targetvpsid'} in preparation for the restore. \n");
                    system("/usr/sbin/vzctl destroy $Backup->{'targetvpsid'}");
                }
                # Copy the data from the backup location to the production environment:
                &debug_msg("INFO: Starting local RSYNC from $Backup->{'storage'}/ to /vz/private/$Backup->{'targetvpsid'}/ to restore VPS $Backup->{'targetvpsid'} \n");
                system("/usr/bin/rsync -q -ar $Backup->{'storage'}/ /vz/private/$Backup->{'targetvpsid'}/ --delete-after");
                &debug_msg("INFO: Copying VPS config from $vps_config_file to /etc/vz/conf/$Backup->{'targetvpsid'}.conf \n");
                system("/bin/cp $vps_config_file /etc/vz/conf/$Backup->{'targetvpsid'}.conf");
                system("/bin/chmod 644 /etc/vz/conf/$Backup->{'targetvpsid'}.conf");
                # Start the VPS again:
                &debug_msg("INFO: Restore of VPS $Backup->{'targetvpsid'} finished. Starting the restored VPS now. \n");
                system("/usr/sbin/vzctl start $Backup->{'targetvpsid'}");
            }
            #
            ### Local Backup + Ploop:
            #
            elsif ((-d "$Backup->{'storage'}" ) && (-e "$Backup->{'storage'}/vps.conf" )) {
                # Make sure we have a tarball:
                $check = `ls -k1 $Backup->{'storage'}/vzpbackup*.tar|wc -l`;
                chomp($check);
                if ($check gt "0") {
                    $tarball = `ls -k1 $Backup->{'storage'}/vzpbackup*.tar|head -1`;
                    chomp($tarball);
                    &debug_msg("INFO: tarball location: $tarball \n");
                    if ($tarball ne "") {
                        # Stop and destroy existing VPS:
                        if (($Backup->{'virttype'} eq "openvz") && (-d "/vz/private/$Backup->{'targetvpsid'}")) {
                            &debug_msg("INFO: Stopping old VPS $Backup->{'targetvpsid'} in preparation for the restore. \n");
                            system("/usr/sbin/vzctl stop $Backup->{'targetvpsid'}");
                            &debug_msg("INFO: Destroying old VPS $Backup->{'targetvpsid'} in preparation for the restore. \n");
                            system("/usr/sbin/vzctl destroy $Backup->{'targetvpsid'}");
                        }
                        &debug_msg("Running: /usr/sausalito/sbin/vzprestore.sh --container=$Backup->{'targetvpsid'} --archive=$tarball --delete-backup-snapshot=yes --confirm=no \n");
                        system("/usr/sausalito/sbin/vzprestore.sh --container=$Backup->{'targetvpsid'} --archive=$tarball --delete-backup-snapshot=yes --confirm=no");
                        # Start the VPS again:
                        &debug_msg("INFO: Restore of VPS $Backup->{'targetvpsid'} finished. Starting the restored VPS now. \n");
                        system("/usr/sbin/vzctl start $Backup->{'targetvpsid'}");
                    }
                    else {
                        &debug_msg("ERROR: No tarball found for restore. \n");
                        $cce->warn('[[base-vserver.failRestoreLocal]]');
                        $cce->bye('FAIL');
                        exit(1);
                    }
                }
                else {
                    &debug_msg("ERROR: Did not find a single tarball suitable for restore. \n");
                    $cce->warn('[[base-vserver.failRestoreLocal]]');
                    $cce->bye('FAIL');
                    exit(1);
                }
            }
            else {
                &debug_msg("ERROR: Did not have all details for a restore. \n");
                $cce->warn('[[base-vserver.failRestoreLocal]]');
                $cce->bye('FAIL');
                exit(1);
            }
        }
        else {
            # This is not supposed to happen:
            &debug_msg("WARN: OID $event->{'OID'} containing an *unknown* type of VPS backup ($vpsid) at $Backup->{'storage'} is NOT going to be restored. \n");
        }
    }
    if (($Backup->{'location'} eq "remote") && ($Backup->{'origin'} eq "rsync") && ($Backup->{'targetvpsid'})) {
        $vpsid = $Backup->{'vpsid'};
        # Handle remote restore of OpenVZ VPS's:
        if ($Backup->{'virttype'} eq "openvz") {
            &debug_msg("INFO: OID $event->{'OID'} containing a remote rsync backup at $Backup->{'storage'} is going to be restored. \n");
            $remote_conf_intern = "$remote_location_vzdump/$vpsid/etc/vzdump/vps.conf";
            &debug_msg("INFO: Checking if $remote_conf_intern exists. \n");
            if ($targethost) {
                $target = $user . '@' . $targethost;
                $target_short = $targethost;
            }
            else {   
                $target = $user . '@' . $targetip;
                $target_short = $targetip;
            }

            #
            ### Remote Backup + Simfs:
            #

            $scheck = `ssh -p $port $target_short "ls -la $remote_conf_intern|wc -l"`;
            chomp($scheck);
            if ($scheck eq "1") {
                # OpenVZ container already present at this VPSID? If so, stop and destroy it:
                if (($Backup->{'virttype'} eq "openvz") && (-d "/vz/private/$Backup->{'targetvpsid'}")) {
                    &debug_msg("INFO: Stopping old VPS $Backup->{'targetvpsid'} in preparation for the restore. \n");
                    system("/usr/sbin/vzctl stop $Backup->{'targetvpsid'}");
                    &debug_msg("INFO: Destroying old VPS $Backup->{'targetvpsid'} in preparation for the restore. \n");
                    system("/usr/sbin/vzctl destroy $Backup->{'targetvpsid'}");
                }
                # Copy the data from the remote backup location to the production environment:
                &debug_msg("INFO: Starting remote RSYNC from $target_short:$Backup->{'storage'}/ to /vz/private/$Backup->{'targetvpsid'}/ to restore VPS $Backup->{'targetvpsid'} \n");
                system("/usr/bin/rsync -q -ar -e 'ssh -p $port' $target_short:$Backup->{'storage'}/ /vz/private/$Backup->{'targetvpsid'}/ --delete-after");
                &debug_msg("INFO: Copying VPS config from /vz/private/$Backup->{'targetvpsid'}/etc/vzdump/vps.conf to /etc/vz/conf/$Backup->{'targetvpsid'}.conf \n");
                system("/bin/cp /vz/private/$Backup->{'targetvpsid'}/etc/vzdump/vps.conf /etc/vz/conf/$Backup->{'targetvpsid'}.conf");
                system("/bin/chmod 644 /etc/vz/conf/$Backup->{'targetvpsid'}.conf");
                # Start the VPS again:
                &debug_msg("INFO: Restore of VPS $Backup->{'targetvpsid'} finished. Starting the restored VPS now. \n");
                system("/usr/sbin/vzctl start $Backup->{'targetvpsid'}");
            }

            #
            ### Remote Backup + Ploop:
            #

            &debug_msg("INFO: Checking if tarball exists at $remote_location_vzdump/$vpsid/vzpbackup*.tar \n");
            $pcheck = `ssh -p $port $target_short "ls -la $remote_location_vzdump/$vpsid/vzpbackup*.tar|wc -l"`;
            chomp($pcheck);
            if ($pcheck gt "0") {
                # At least one tarball exists:
                $tarball = `ssh -p $port $target_short "ls -k1 $Backup->{'storage'}/vzpbackup*.tar|head -1"`;
                chomp($tarball);
                &debug_msg("INFO: tarball location: $tarball \n");
                if ($tarball ne "") {
                    # Tarball found. Pull it to local destination:
                    $temp_vps_dump = "/vz/vztmp/" . $Backup->{'targetvpsid'} . '.tar';
                    &debug_msg("INFO: Starting remote RSYNC of Tarball $tarball to $temp_vps_dump to restore VPS $Backup->{'targetvpsid'} \n");
                    &debug_msg("Running: /usr/bin/rsync -q -ar -e 'ssh -p $port' $target_short:$tarball $temp_vps_dump --delete-after \n");
                    system("/usr/bin/rsync -q -ar -e 'ssh -p $port' $target_short:$tarball $temp_vps_dump --delete-after");

                    # Stop and destroy existing VPS:
                    if (($Backup->{'virttype'} eq "openvz") && (-d "/vz/private/$Backup->{'targetvpsid'}")) {
                        &debug_msg("INFO: Stopping old VPS $Backup->{'targetvpsid'} in preparation for the restore. \n");
                        system("/usr/sbin/vzctl stop $Backup->{'targetvpsid'}");
                        &debug_msg("INFO: Destroying old VPS $Backup->{'targetvpsid'} in preparation for the restore. \n");
                        system("/usr/sbin/vzctl destroy $Backup->{'targetvpsid'}");
                    }
                    &debug_msg("Running: /usr/sausalito/sbin/vzprestore.sh --container=$Backup->{'targetvpsid'} --archive=$temp_vps_dump --delete-backup-snapshot=yes --confirm=no \n");
                    system("/usr/sausalito/sbin/vzprestore.sh --container=$Backup->{'targetvpsid'} --archive=$temp_vps_dump --delete-backup-snapshot=yes --confirm=no");
                    system("rm -f $temp_vps_dump");
                    # Start the VPS again:
                    &debug_msg("INFO: Restore of VPS $Backup->{'targetvpsid'} finished. Starting the restored VPS now. \n");
                    system("/usr/sbin/vzctl start $Backup->{'targetvpsid'}");
                }
                else {
                    &debug_msg("ERROR: Unable to determine name of remote backup tarball! \n");
                    $cce->warn('[[base-vserver.failRestoreRemote]]');
                    $cce->bye('FAIL');
                    exit(1);
                }
            }

            # Found neither Simfs or Ploop backup in expected formats:
            if (($scheck eq "0") && ($pcheck eq "0")) {
                &debug_msg("ERROR: You know nothing, Jon Snow! (Neither Simfs nor Ploop backup found) \n");
                $cce->warn('[[base-vserver.failRestoreRemote]]');
                $cce->bye('FAIL');
                exit(1);
            }
        }
        else {
            # This is not supposed to happen:
            &debug_msg("WARN: OID $event->{'OID'} containing an *unknown* type of remote VPS backup ($vpsid) at $Backup->{'storage'} is NOT going to be restored. \n");
        }
    }
}

$cce->bye('SUCCESS');
exit 0;

# Debugger:
sub debug_msg {
    if ($DEBUG) {
        my $msg = shift;
        $user = $ENV{'USER'};
        setlogsock('unix');
        openlog($0,'','user');
        syslog('info', "$ARGV[0]: $msg");
        closelog;
    }
}

# 
# Copyright (c) 2016 Michael Stauber, SOLARSPEED.NET
# All Rights Reserved.
# 
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in 
#    the documentation and/or other materials provided with the 
#    distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its 
#    contributors may be used to endorse or promote products derived 
#    from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
# 
# You acknowledge that this software is not designed or intended for 
# use in the design, construction, operation or maintenance of any 
# nuclear facility.
# 